"""app/core/*"""

from textwrap import dedent

APP_CORE_INIT_CONTENT = dedent(
    """
from .config import settings
from .lifespan import lifespan

__all__ = [settings, lifespan]

"""
)


APP_CORE_CONFIG_CONTENT = dedent("""
from pydantic_settings import BaseSettings, SettingsConfigDict
from dotenv import load_dotenv
from pydantic import Field
from functools import lru_cache
from typing import Literal

load_dotenv()


class Settings(BaseSettings):
    '''
    Settings for the application
    '''

    model_config = SettingsConfigDict(
        env_file=".env", env_file_encoding="utf-8", extra="ignore"
    )

    TITLE: str = Field(
        description="The title of the application", validation_alias="TITLE"
    )

    DESCRIPTION: str = Field(
        description="The description of the application",
        validation_alias="DESCRIPTION",
    )

    VERSION: str = Field(
        description="The version of the application", validation_alias="VERSION"
    )

    DEBUG: bool = Field(
        description="Whether the application is in debug mode",
        validation_alias="DEBUG",
    )

    PORT: int = Field(
        description="The port to run the application on", validation_alias="PORT"
    )

    HOST: str = Field(
        description="The host to run the application on", validation_alias="HOST"
    )

    DB_USER: str = Field(..., description="Database user name", validation_alias="DB_USER")

    DB_PASS: str = Field(..., description="Database password", validation_alias="DB_PASS")

    DB_HOST: str = Field(..., description="Database Host", validation_alias="DB_HOST")

    DB_PORT: str = Field(..., description="Database port", validation_alias="DB_PORT")

    DB_NAME: str = Field(..., description="Database Name", validation_alias="DB_NAME")

    DB_TYPE: Literal["postgres", "mysql", "sqlite", "mongodb"] = Field(
        "postgres", description="Which database to use for the project", validation_alias="DB_TYPE"
    )

    @property
    def db_url(self) -> str:
        if self.DB_TYPE == "postgres":
            return f"postgresql+asyncpg://{self.DB_USER}:{self.DB_PASS}@{self.DB_HOST}:{self.DB_PORT}/{self.DB_NAME}"


@lru_cache(maxsize=1)
def get_settings() -> Settings:
    '''
    Get the settings
    '''
    return Settings()


settings = get_settings()
""")


APP_CORE_LIFESPAN_CONTENT = dedent("""
from fastapi import FastAPI
from contextlib import asynccontextmanager
from app.database.session import engine


@asynccontextmanager
async def lifespan(app: FastAPI):
    yield
    await engine.dispose()
""")



from textwrap import dedent

ENV_FILE_CONTENT = dedent("""
# Project
TITLE="{title}"
DESCRIPTION="{description}"
VERSION="{version}"

# Debugging
DEBUG=True

# Server
PORT=8000
HOST="0.0.0.0"

# Database
DB_USER="postgres"
DB_PASS="postgres"
DB_HOST="loclahost"
DB_PORT=5432
DB_NAME="postgres"
DB_TYPE="{DB_TYPE}"
""")



from typing import Literal
from builders_hut.utils import make_file, write_file, run_subprocess, get_python_file
from pathlib import Path
from builders_hut.setups.file_contents import APP_DATABASE_SESSION_CONTENT_FOR_SQL


class DatabaseFactory:
    def __init__(
        self,
        database_type: Literal["sql", "nosql"],
        location: Path,
    ):
        self.database_type = database_type
        self.location = location

    def setup_db(self):
        match self.database_type:
            case "sql":
                make_file(self.location / "app" / "database" / "session.py")
                write_file(
                    self.location / "app" / "database" / "session.py",
                    APP_DATABASE_SESSION_CONTENT_FOR_SQL,
                )

                command = "alembic init -t async migrations"

                python_file = get_python_file()

                full_command = f"{python_file} {command}"

                run_subprocess(self.location, full_command)
            case "nosql":
                pass
            case _:
                raise RuntimeError("Invalid Database Type Selected")



"""app/models/*"""

from textwrap import dedent

APP_MODELS_COMMON_CONTENT = dedent("""
from sqlmodel import Field, SQLModel
from datetime import datetime, timezone
import uuid


class BaseModel(SQLModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)

    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        nullable=False,
    )

    updated_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        nullable=False,
    )
""")



from .app_main import APP_MAIN_CONTENT
from .app_api import APP_API_INIT_CONTENT, APP_API_COMMON_CONTENT
from .app_core import (
    APP_CORE_INIT_CONTENT,
    APP_CORE_CONFIG_CONTENT,
    APP_CORE_LIFESPAN_CONTENT,
)
from .app_database import APP_DATABASE_SESSION_CONTENT_FOR_SQL
from .app_model import APP_MODELS_COMMON_CONTENT
from .app_templates import APP_TEMPLATES_INDEX_CONTENT


from .env_file import ENV_FILE_CONTENT
from .gitignore_file import GIT_IGNORE_FILE_CONTENT


__all__ = [
    # main
    APP_MAIN_CONTENT,
    # API
    APP_API_INIT_CONTENT,
    APP_API_COMMON_CONTENT,
    # Core
    APP_CORE_INIT_CONTENT,
    APP_CORE_CONFIG_CONTENT,
    APP_CORE_LIFESPAN_CONTENT,
    # Database
    APP_DATABASE_SESSION_CONTENT_FOR_SQL,
    # Model
    APP_MODELS_COMMON_CONTENT,
    # Templates
    APP_TEMPLATES_INDEX_CONTENT,
    # .env File
    ENV_FILE_CONTENT,
    # .gitignore File
    GIT_IGNORE_FILE_CONTENT,
]



from builders_hut.setups import BaseSetup
from typing import Literal
from builders_hut.setups.database import DatabaseFactory


class SetupDatabase(BaseSetup):
    """setup database"""

    def create(self):
        DatabaseFactory(self.database_type, self.location).setup_db()

    def configure(self, database_type: Literal["sql", "nosql"], **kwargs):
        self.database_type = database_type



"app/main.py"

from textwrap import dedent

APP_MAIN_CONTENT = dedent("""
from fastapi import FastAPI
from app.core import settings, lifespan
from app.api import common_router


def create_app() -> FastAPI:
    '''
    Create a fastapi app
    '''
    app = FastAPI(
        docs_url=None,
        redoc_url=None,
        title=settings.TITLE,
        description=settings.DESCRIPTION,
        version=settings.VERSION,
        lifespan=lifespan,
    )

    app.include_router(common_router)

    return app


app = create_app()
""")



from builders_hut.setups import FILES_TO_WRITE, BaseSetup
from builders_hut.utils import write_file


class SetupFileWriter(BaseSetup):
    """
    Write data to the files created previously
    """

    def create(self):
        self._write_files()

    def _write_files(self) -> None:
        for path, content in FILES_TO_WRITE.items():
            if path.name == ".env":
                content = content.format(
                    title=self.name,
                    description=self.description,
                    version=self.version,
                    DB_TYPE=self.database_provider,
                )
            path = self.location / path
            write_file(path, content)

    def configure(
        self,
        name: str,
        description: str,
        version: str,
        database_provider: str,
        **kwargs,
    ):
        self.name = name
        self.description = description
        self.version = version
        self.database_provider = database_provider



from textwrap import dedent

GIT_IGNORE_FILE_CONTENT = dedent("""
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class

# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST

# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/

# Celery stuff
celerybeat-schedule
celerybeat.pid

# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/


# mypy
.mypy_cache/
.dmypy.json
dmypy.json

# Pyre type checker
.pyre/

# pytype static type analyzer
.pytype/


# ruff
.ruff_cache/
""")



"app/database/*"

from textwrap import dedent

APP_DATABASE_SESSION_CONTENT_FOR_SQL = dedent('''
"""
Database Connections Management
"""

from typing import AsyncGenerator, Annotated

from fastapi import Depends
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    create_async_engine,
    async_sessionmaker,
)

from app.core import settings


""" database engine """
engine: AsyncEngine = create_async_engine(
    settings.db_url,
    echo=False,
    future=True,
    pool_pre_ping=True,
)

""" databse session """
asyncsession = async_sessionmaker(
    bind=engine,
    class_=AsyncSession,
    expire_on_commit=False,
)


async def get_session() -> AsyncGenerator[AsyncSession, None]:
    """get the session as needed"""
    async with asyncsession() as session:
        yield session


""" session dependency to use when needed """
SessionDeps = Annotated[AsyncSession, Depends(get_session)]
''')



from pathlib import Path

import typer
from rich.console import Console
from typing import Literal
from builders_hut.setups import (
    SetupEnv,
    SetupFiles,
    SetupFileWriter,
    SetupStructure,
    SetupDatabase,
    SetupGithub,
)
from builders_hut.utils import setup_project
from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    BarColumn,
    TimeElapsedColumn,
)


console = Console()
app = typer.Typer(
    invoke_without_command=True,
    no_args_is_help=True,
)

BANNER = r"""
[bold cyan]
 ____        _ _     _                 _   _       _   
| __ ) _   _(_) | __| | ___ _ __ ___  | | | |_   _| |_ 
|  _ \| | | | | |/ _` |/ _ \ '__/ __| | |_| | | | | __|
| |_) | |_| | | | (_| |  __/ |  \__ \ |  _  | |_| | |_ 
|____/ \__,_|_|_|\__,_|\___|_|  |___/ |_| |_|\__,_|\__|
[/bold cyan]
"""

APP_VERSION = "0.4.0"


@app.callback()
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        False,
        "--version",
        "-v",
        help="Show the version and exit",
        is_eager=True,
    ),
):
    """Builders Hut – FastAPI Scaffolding Tool"""
    console.print(BANNER)

    if version:
        console.print(f"[bold green]hut version {APP_VERSION}[/bold green]")

    if ctx.invoked_subcommand is None:
        typer.echo(ctx.get_help())
        raise typer.Exit()


@app.command()
def build(
    name: str = typer.Option(
        Path.cwd().name,
        "--name",
        "-n",
        help="Project name",
    ),
    description: str = typer.Option(
        "A new project",
        "--description",
        "-d",
        help="Project description",
    ),
    version: str = typer.Option(
        "0.1.0",
        "--version",
        "-v",
        help="Project version",
    ),
    path: Path = typer.Option(
        Path.cwd(),
        "--path",
        "-p",
        help="Project directory",
    ),
    database_type: Literal["sql", "nosql"] = typer.Option(
        "sql",
        "--database-type",
        "-dt",
        help="Database Type (sql, nosql)",
    ),
    database_provider: Literal["postgres", "mysql", "sqlite", "mongodb"] = typer.Option(
        "postgres",
        "--database-provider",
        "-dp",
        help="Database Provider (postgres, mysql, sqlite, mongodb)",
    ),
    accept_default: bool = typer.Option(
        None,
        "--accept_defaults",
        "-y",
        help="Run command with all default values selected",
    ),
):
    """
    Build project structure (CLI or interactive).
    """
    # ===== Defaults =====
    DEFAULTS = {
        "name": Path.cwd().name,
        "description": "A new project",
        "version": "0.1.0",
        "database_type": "sql",
        "database_provider": "postgres",
    }

    # ===== Apply defaults or prompt =====
    if accept_default:
        name = name or DEFAULTS["name"]
        description = description or DEFAULTS["description"]
        version = version or DEFAULTS["version"]
        database_type = database_type or DEFAULTS["database_type"]
        database_provider = database_provider or DEFAULTS["database_provider"]
    else:
        name = typer.prompt("Enter Project Title", default=DEFAULTS["name"])
        description = typer.prompt(
            "Enter Project Description", default=DEFAULTS["description"]
        )
        version = typer.prompt("Enter Project Version", default=DEFAULTS["version"])
        database_type = typer.prompt(
            "Enter Database Type", default=DEFAULTS["database_type"]
        )
        database_provider = typer.prompt(
            "Enter Database Provider", default=DEFAULTS["database_provider"]
        )

    try:
        project_location = path.resolve()

        setup_steps = [
            SetupStructure,
            SetupFiles,
            SetupGithub,
            SetupEnv,
            SetupFileWriter,
            SetupDatabase,
        ]

        with Progress(
            SpinnerColumn(),
            TextColumn("[bold cyan]{task.description}"),
            BarColumn(),
            TimeElapsedColumn(),
        ) as progress:
            task = progress.add_task("Setting up project", total=len(setup_steps))

            for setup_cls in setup_steps:
                progress.update(task, description=f"Running {setup_cls.__name__}")

                setup_project(
                    project_location,
                    setup_cls,
                    name=name,
                    description=description,
                    version=version,
                    database_type=database_type,
                    database_provider=database_provider,
                )

                progress.advance(task)

        typer.secho(
            "Project setup completed successfully.",
            fg=typer.colors.GREEN,
        )
        typer.secho(
            "Update .env File Content Before Starting The Server",
            fg=typer.colors.MAGENTA,
        )

    except Exception as e:
        typer.secho(
            f"Project setup failed: {e}",
            fg=typer.colors.RED,
        )
        raise typer.Exit(code=1)


@app.command()
def add():
    """Add components to an existing project."""
    console.print("[yellow]Add command not implemented yet[/yellow]")
